use std::collections::HashMap;
use std::process::Stdio;

use async_trait::async_trait;
use serde::{Deserialize, Serialize};
use tokio::io::{AsyncBufReadExt, AsyncWriteExt, BufReader, BufWriter};
use tokio::process::{Child, ChildStdin, ChildStdout, Command};

use super::{CopyDirection, CopyOptions, Host, RebootOptions};
use crate::error::{ColmenaError, ColmenaResult};
use crate::job::JobHandle;
use crate::nix::{self, Key, Profile, StorePath, SystemTypeConfig};

pub type TransportId = String;

#[derive(Clone, Debug, PartialEq, Eq, Serialize, Deserialize)]
pub struct Transport {
    id: TransportId,
    long_name: String,
    description: String,
}

pub type GoalId = String;

#[derive(Clone, Debug, PartialEq, Eq, Serialize, Deserialize)]
pub struct Goal {
    id: GoalId,
    long_name: String,
    description: String,
}

/// A request to the activation program
#[derive(Clone, Debug, Serialize, Deserialize)]
pub enum Request {
    /// Asks for activation program's capabilities
    Capabilities,

    Connection {
        connection_options: HashMap<String, String>,
    },

    /// Copy closure to/from host
    CopyClosure {
        path: StorePath,
        to_host: bool,
        options: CopyOptions,
    },

    /// Deploys the profile to host
    Deploy {
        goal: GoalId,
        toplevel: StorePath,
        options: CopyOptions,
    },

    /// Realizes the derivation
    Realize {
        path: StorePath,
    },

    /// Uploads keys to host
    UploadKeys {
        keys: HashMap<String, Key>,
        require_ownership: bool,
    },

    Activate {
        profile: StorePath,
        goal: GoalId,
    },

    Reboot {
        wait_for_boot: bool,
    },
}

#[derive(Clone, Debug, PartialEq, Eq, Serialize, Deserialize)]
pub enum CapabilityResponse {
    V1 {
        supported_transports: Vec<Transport>,
        supported_goals: Vec<Goal>,
    },
}

/// A response from the activation program
#[derive(Clone, Debug, PartialEq, Eq, Serialize, Deserialize)]
pub enum Response {
    Capabilities(CapabilityResponse),
    Progress { phase: String },
    NewStorePath { store_path: StorePath },
    Failed { error: String },
}

#[derive(Debug)]
pub struct OpenedCommand {
    command: Child,
    stdin: BufWriter<ChildStdin>,
    stdout: BufReader<ChildStdout>,
}

#[derive(Debug)]
pub struct GenericHost {
    activation_program: OpenedCommand,
    connection_options: HashMap<String, String>,
}

#[async_trait]
impl Host for GenericHost {
    async fn copy_closure(
        &mut self,
        closure: &StorePath,
        direction: CopyDirection,
        options: CopyOptions,
    ) -> ColmenaResult<()> {
        self.call_default_handler(Request::CopyClosure {
            path: closure.clone(),
            to_host: direction == CopyDirection::ToRemote,
            options,
        })
        .await
    }

    async fn realize_remote(&mut self, derivation: &StorePath) -> ColmenaResult<Vec<StorePath>> {
        Ok(self
            .call(
                Request::Realize {
                    path: derivation.clone(),
                },
                move |response, mut store_paths| {
                    match response {
                        Response::Progress { phase } => println!("{}", phase),
                        Response::NewStorePath { store_path } => store_paths.push(store_path),
                        _ => (),
                    };
                    store_paths
                },
                Vec::new(),
            )
            .await?)
    }

    fn set_job(&mut self, _: Option<JobHandle>) {}

    async fn deploy(
        &mut self,
        profile: &Profile,
        goal: nix::Goal,
        copy_options: CopyOptions,
    ) -> ColmenaResult<()> {
        self.call_default_handler(Request::Deploy {
            goal: goal.to_string(),
            toplevel: profile.as_store_path().clone(),
            options: copy_options,
        })
        .await
    }

    async fn upload_keys(
        &mut self,
        keys: &HashMap<String, Key>,
        require_ownership: bool,
    ) -> ColmenaResult<()> {
        self.call_default_handler(Request::UploadKeys {
            keys: keys.clone(),
            require_ownership,
        })
        .await
    }

    async fn get_current_system_profile(&mut self) -> ColmenaResult<Profile> {
        Err(ColmenaError::Unsupported)
    }

    async fn get_main_system_profile(&mut self) -> ColmenaResult<Profile> {
        Err(ColmenaError::Unsupported)
    }

    async fn activate(&mut self, profile: &Profile, goal: nix::Goal) -> ColmenaResult<()> {
        self.call_default_handler(Request::Activate {
            profile: profile.as_store_path().clone(),
            goal: goal.to_string(),
        })
        .await
    }

    async fn reboot(&mut self, options: RebootOptions) -> ColmenaResult<()> {
        self.call_default_handler(Request::Reboot {
            wait_for_boot: options.wait_for_boot,
        })
        .await
    }
}

impl GenericHost {
    pub fn new(
        system: &SystemTypeConfig,
        connection_options: &HashMap<String, String>,
    ) -> ColmenaResult<GenericHost> {
        let mut command = Command::new(system.activation_program.as_path())
            .stdin(Stdio::piped())
            .stdout(Stdio::piped())
            .stderr(Stdio::inherit())
            .spawn()?;

        let stdin = BufWriter::new(command.stdin.take().unwrap());
        let stdout = BufReader::new(command.stdout.take().unwrap());

        let activation_program = OpenedCommand {
            command,
            stdin,
            stdout,
        };

        Ok(Self {
            activation_program,
            connection_options: connection_options.clone(),
        })
    }

    pub async fn connect(&mut self) -> ColmenaResult<()> {
        self.call_default_handler(Request::Connection {
            connection_options: self.connection_options.clone(),
        })
        .await
    }

    async fn call<F, T>(
        &mut self,
        request: Request,
        handler: F,
        initial_value: T,
    ) -> ColmenaResult<T>
    where
        F: Fn(Response, T) -> T,
    {
        let json = serde_json::to_string(&request)
            .map_err(|error| ColmenaError::InvalidActivationProgramJson { error })?;

        log::trace!("giving to activation program stdin {}", json.as_str());

        let stdin = &mut self.activation_program.stdin;
        let stdout = &mut self.activation_program.stdout;

        stdin.write_all(json.as_bytes()).await?;
        stdin.write_all(b"\n").await?;
        stdin.flush().await?;

        let mut line = String::new();
        let mut value = initial_value;

        while stdout.read_line(&mut line).await.is_ok_and(|x| x != 0) {
            if line == "\n" {
                break;
            }
            log::trace!("receiving from activation program:\n{}", line);
            let response: Response = serde_json::from_str(line.as_str())
                .map_err(|error| ColmenaError::InvalidActivationProgramJson { error })?;
            value = handler(response, value);
        }

        Ok(value)
    }

    async fn call_default_handler(&mut self, request: Request) -> ColmenaResult<()> {
        self.call(
            request,
            |response, _| match response {
                Response::Progress { phase } => println!("{phase}"),
                _ => (),
            },
            (),
        )
        .await
    }
}
